@@ -22,17 +22,19 @@ import math
 import os
 import pickle
 import signal
+import threading
 import warnings
 from collections import defaultdict
 from datetime import datetime, timedelta
 from functools import partial
+from inspect import currentframe
 from tempfile import NamedTemporaryFile
-from typing import IO, TYPE_CHECKING, Any, Callable, Dict, Iterable, List, NamedTuple, Optional, Tuple, Union
+from types import TracebackType
+from typing import IO, TYPE_CHECKING, Any, Iterable, List, NamedTuple, Optional, Tuple, Union
 from urllib.parse import quote
 
 import dill
 import jinja2
-import lazy_object_proxy
 import pendulum
 from jinja2 import TemplateAssertionError, UndefinedError
 from sqlalchemy import (
@@ -62,7 +64,6 @@ from airflow.configuration import conf
 from airflow.exceptions import (
     AirflowException,
     AirflowFailException,
-    AirflowNotFoundException,
     AirflowRescheduleException,
     AirflowSensorTimeout,
     AirflowSkipException,
@@ -73,12 +74,10 @@ from airflow.exceptions import (
     TaskDeferred,
 )
 from airflow.models.base import COLLATION_ARGS, ID_LEN, Base
-from airflow.models.connection import Connection
 from airflow.models.log import Log
 from airflow.models.param import ParamsDict
 from airflow.models.taskfail import TaskFail
 from airflow.models.taskreschedule import TaskReschedule
-from airflow.models.variable import Variable
 from airflow.models.xcom import XCOM_RETURN_KEY, XCom
 from airflow.plugins_manager import integrate_macros_plugins
 from airflow.sentry import Sentry
@@ -88,15 +87,17 @@ from airflow.ti_deps.dependencies_deps import REQUEUEABLE_DEPS, RUNNING_DEPS
 from airflow.timetables.base import DataInterval
 from airflow.typing_compat import Literal
 from airflow.utils import timezone
+from airflow.utils.context import ConnectionAccessor, Context, VariableAccessor
 from airflow.utils.email import send_email
-from airflow.utils.helpers import is_container
+from airflow.utils.helpers import is_container, render_template_to_string
 from airflow.utils.log.logging_mixin import LoggingMixin
 from airflow.utils.net import get_hostname
 from airflow.utils.operator_helpers import context_to_airflow_vars
 from airflow.utils.platform import getuser
+from airflow.utils.retries import run_with_db_retries
 from airflow.utils.session import create_session, provide_session
-from airflow.utils.sqlalchemy import ExtendedJSON, UtcDateTime
-from airflow.utils.state import DagRunState, State
+from airflow.utils.sqlalchemy import ExtendedJSON, UtcDateTime, with_row_locks
+from airflow.utils.state import DagRunState, State, TaskInstanceState
 from airflow.utils.timeout import timeout
 
 try:
@@ -108,18 +109,20 @@ except ImportError:
     ApiClient = None
 
 TR = TaskReschedule
-Context = Dict[str, Any]
 
 _CURRENT_CONTEXT: List[Context] = []
 log = logging.getLogger(__name__)
 
 
 if TYPE_CHECKING:
+    from airflow.models.baseoperator import BaseOperator
     from airflow.models.dag import DAG, DagModel, DagRun
 
+_TASK_EXECUTION_FRAME_LOCAL_STORAGE = threading.local()
+
 
 @contextlib.contextmanager
-def set_current_context(context: Context):
+def set_current_context(context: Context) -> None:
     """
     Sets the current execution context to the provided context object.
     This method should be called once per Task execution, before calling operator.execute.
@@ -183,11 +186,11 @@ def clear_task_instances(
     job_ids = []
     task_id_by_key = defaultdict(lambda: defaultdict(lambda: defaultdict(set)))
     for ti in tis:
-        if ti.state == State.RUNNING:
+        if ti.state == TaskInstanceState.RUNNING:
             if ti.job_id:
                 # If a task is cleared when running, set its state to RESTARTING so that
                 # the task is terminated and becomes eligible for retry.
-                ti.state = State.RESTARTING
+                ti.state = TaskInstanceState.RESTARTING
                 job_ids.append(ti.job_id)
         else:
             task_id = ti.task_id
@@ -202,7 +205,7 @@ def clear_task_instances(
                 # outdated. We make max_tries the maximum value of its
                 # original max_tries or the last attempted try number.
                 ti.max_tries = max(ti.max_tries, ti.prev_attempted_tries)
-            ti.state = State.NONE
+            ti.state = None
             ti.external_executor_id = None
             session.merge(ti)
 
@@ -240,7 +243,7 @@ def clear_task_instances(
         from airflow.jobs.base_job import BaseJob
 
         for job in session.query(BaseJob).filter(BaseJob.id.in_(job_ids)).all():
-            job.state = State.RESTARTING
+            job.state = TaskInstanceState.RESTARTING
 
     if activate_dag_runs is not None:
         warnings.warn(
@@ -269,10 +272,11 @@ def clear_task_instances(
             )
             .all()
         )
+        dag_run_state = DagRunState(dag_run_state)  # Validate the state value.
         for dr in drs:
             dr.state = dag_run_state
             dr.start_date = timezone.utcnow()
-            if dag_run_state == State.QUEUED:
+            if dag_run_state == DagRunState.QUEUED:
                 dr.last_scheduling_decision = None
                 dr.start_date = None
 
@@ -449,6 +453,7 @@ class TaskInstance(Base, LoggingMixin):
         self.run_id = run_id
 
         self.try_number = 0
+        self.max_tries = self.task.retries
         self.unixname = getuser()
         if state:
             self.state = state
@@ -726,7 +731,9 @@ class TaskInstance(Base, LoggingMixin):
         )
 
         if lock_for_update:
-            ti: Optional[TaskInstance] = qry.with_for_update().first()
+            for attempt in run_with_db_retries(logger=self.log):
+                with attempt:
+                    ti: Optional[TaskInstance] = qry.with_for_update().first()
         else:
             ti = qry.first()
         if ti:
@@ -775,7 +782,8 @@ class TaskInstance(Base, LoggingMixin):
         self.pool_slots = task.pool_slots
         self.priority_weight = task.priority_weight_total
         self.run_as_user = task.run_as_user
-        self.max_tries = task.retries
+        # Do not set max_tries to task.retries here because max_tries is a cumulative
+        # value that needs to be stored in the db.
         self.executor_config = task.executor_config
         self.operator = task.task_type
 
@@ -1407,7 +1415,18 @@ class TaskInstance(Base, LoggingMixin):
         """Prepare Task for Execution"""
         from airflow.models.renderedtifields import RenderedTaskInstanceFields
 
+        parent_pid = os.getpid()
+
         def signal_handler(signum, frame):
+            pid = os.getpid()
+
+            # If a task forks during execution (from DAG code) for whatever
+            # reason, we want to make sure that we react to the signal only in
+            # the process that we've spawned ourselves (referred to here as the
+            # parent process).
+            if pid != parent_pid:
+                os._exit(1)
+                return
             self.log.error("Received SIGTERM. Terminating subprocesses.")
             self.task.on_kill()
             raise AirflowException("Task received SIGTERM signal")
@@ -1478,40 +1497,48 @@ class TaskInstance(Base, LoggingMixin):
         # If the task has been deferred and is being executed due to a trigger,
         # then we need to pick the right method to come back to, otherwise
         # we go for the default execute
-        execute_callable = task_copy.execute
         if self.next_method:
             # __fail__ is a special signal value for next_method that indicates
             # this task was scheduled specifically to fail.
             if self.next_method == "__fail__":
                 next_kwargs = self.next_kwargs or {}
+                traceback = self.next_kwargs.get("traceback")
+                if traceback is not None:
+                    self.log.error("Trigger failed:\n%s", "\n".join(traceback))
                 raise TaskDeferralError(next_kwargs.get("error", "Unknown"))
             # Grab the callable off the Operator/Task and add in any kwargs
             execute_callable = getattr(task_copy, self.next_method)
             if self.next_kwargs:
                 execute_callable = partial(execute_callable, **self.next_kwargs)
+        else:
+            execute_callable = task_copy.execute
         # If a timeout is specified for the task, make it fail
         # if it goes beyond
-        if task_copy.execution_timeout:
-            # If we are coming in with a next_method (i.e. from a deferral),
-            # calculate the timeout from our start_date.
-            if self.next_method:
-                timeout_seconds = (
-                    task_copy.execution_timeout - (timezone.utcnow() - self.start_date)
-                ).total_seconds()
+        try:
+            if task_copy.execution_timeout:
+                # If we are coming in with a next_method (i.e. from a deferral),
+                # calculate the timeout from our start_date.
+                if self.next_method:
+                    timeout_seconds = (
+                        task_copy.execution_timeout - (timezone.utcnow() - self.start_date)
+                    ).total_seconds()
+                else:
+                    timeout_seconds = task_copy.execution_timeout.total_seconds()
+                try:
+                    # It's possible we're already timed out, so fast-fail if true
+                    if timeout_seconds <= 0:
+                        raise AirflowTaskTimeout()
+                    # Run task in timeout wrapper
+                    with timeout(timeout_seconds):
+                        result = execute_callable(context=context)
+                except AirflowTaskTimeout:
+                    task_copy.on_kill()
+                    raise
             else:
-                timeout_seconds = task_copy.execution_timeout.total_seconds()
-            try:
-                # It's possible we're already timed out, so fast-fail if true
-                if timeout_seconds <= 0:
-                    raise AirflowTaskTimeout()
-                # Run task in timeout wrapper
-                with timeout(timeout_seconds):
-                    result = execute_callable(context=context)
-            except AirflowTaskTimeout:
-                task_copy.on_kill()
-                raise
-        else:
-            result = execute_callable(context=context)
+                result = execute_callable(context=context)
+        except:  # noqa: E722
+            _TASK_EXECUTION_FRAME_LOCAL_STORAGE.frame = currentframe()
+            raise
         # If the task returns a result, push an XCom containing it
         if task_copy.do_xcom_push and result is not None:
             self.xcom_push(key=XCOM_RETURN_KEY, value=result)
@@ -1655,11 +1682,24 @@ class TaskInstance(Base, LoggingMixin):
         # Don't record reschedule request in test mode
         if test_mode:
             return
+
+        from airflow.models.dagrun import DagRun  # Avoid circular import
+
         self.refresh_from_db(session)
 
         self.end_date = timezone.utcnow()
         self.set_duration()
 
+        # Lock DAG run to be sure not to get into a deadlock situation when trying to insert
+        # TaskReschedule which apparently also creates lock on corresponding DagRun entity
+        with_row_locks(
+            session.query(DagRun).filter_by(
+                dag_id=self.dag_id,
+                run_id=self.run_id,
+            ),
+            session=session,
+        ).one()
+
         # Log reschedule request
         session.add(
             TaskReschedule(
@@ -1685,6 +1725,36 @@ class TaskInstance(Base, LoggingMixin):
         session.commit()
         self.log.info('Rescheduling task, marking task as UP_FOR_RESCHEDULE')
 
+    def get_truncated_error_traceback(self, error: BaseException) -> Optional[TracebackType]:
+        """
+        Returns truncated error traceback.
+
+        This method returns traceback of the error truncated to the
+        frame saved by earlier try/except along the way. If the frame
+        is found, the traceback will be truncated to below the frame.
+
+        :param error: exception to get traceback from
+        :return: traceback to print
+        """
+        tb = error.__traceback__
+        try:
+            execution_frame = _TASK_EXECUTION_FRAME_LOCAL_STORAGE.frame
+        except AttributeError:
+            self.log.warning(
+                "We expected to get frame set in local storage but it was not."
+                " Please report this as an issue with full logs"
+                " at https://github.com/apache/airflow/issues/new",
+                exc_info=True,
+            )
+            return tb
+        _TASK_EXECUTION_FRAME_LOCAL_STORAGE.frame = None
+        while tb is not None:
+            if tb.tb_frame is execution_frame:
+                tb = tb.tb_next
+                break
+            tb = tb.tb_next
+        return tb or error.__traceback__
+
     @provide_session
     def handle_failure(
         self,
@@ -1700,7 +1770,8 @@ class TaskInstance(Base, LoggingMixin):
 
         if error:
             if isinstance(error, Exception):
-                self.log.error("Task failed with exception", exc_info=error)
+                tb = self.get_truncated_error_traceback(error)
+                self.log.error("Task failed with exception", exc_info=(type(error), error, tb))
             else:
                 self.log.error("%s", error)
             # external monitoring process provides pickle file so _run_raw_task
@@ -1782,23 +1853,25 @@ class TaskInstance(Base, LoggingMixin):
         # Do not use provide_session here -- it expunges everything on exit!
         if not session:
             session = settings.Session()
-        task = self.task
-        dag: DAG = task.dag
+
         from airflow import macros
 
         integrate_macros_plugins()
 
+        task: "BaseOperator" = self.task
+        dag: DAG = task.dag
         dag_run = self.get_dagrun(session)
         data_interval = dag.get_run_data_interval(dag_run)
 
+        # Validates Params and convert them into a simple dict.
         params = ParamsDict(suppress_exception=ignore_param_exceptions)
-
         with contextlib.suppress(AttributeError):
             params.update(dag.params)
         if task.params:
             params.update(task.params)
         if conf.getboolean('core', 'dag_run_conf_overrides_params'):
             self.overwrite_params_with_dag_run_conf(params=params, dag_run=dag_run)
+        task.params = params.validate()
 
         logical_date = timezone.coerce_datetime(self.execution_date)
         ds = logical_date.strftime('%Y-%m-%d')
@@ -1807,9 +1880,6 @@ class TaskInstance(Base, LoggingMixin):
         ts_nodash = logical_date.strftime('%Y%m%dT%H%M%S')
         ts_nodash_with_tz = ts.replace('-', '').replace(':', '')
 
-        # Now validates Params and convert them into a simple dict
-        task.params = params.validate()
-
         @cache  # Prevent multiple database access.
         def _get_previous_dagrun_success() -> Optional["DagRun"]:
             return self.get_previous_dagrun(state=State.SUCCESS, session=session)
@@ -1838,121 +1908,16 @@ class TaskInstance(Base, LoggingMixin):
                 return None
             return timezone.coerce_datetime(dagrun.start_date)
 
-        # Custom accessors.
-        class VariableAccessor:
-            """
-            Wrapper around Variable. This way you can get variables in
-            templates by using ``{{ var.value.variable_name }}`` or
-            ``{{ var.value.get('variable_name', 'fallback') }}``.
-            """
-
-            def __init__(self):
-                self.var = None
-
-            def __getattr__(
-                self,
-                item: str,
-            ):
-                self.var = Variable.get(item)
-                return self.var
-
-            def __repr__(self):
-                return str(self.var)
-
-            @staticmethod
-            def get(
-                item: str,
-                default_var: Any = Variable._Variable__NO_DEFAULT_SENTINEL,
-            ):
-                """Get Airflow Variable value"""
-                return Variable.get(item, default_var=default_var)
-
-        class VariableJsonAccessor:
-            """
-            Wrapper around Variable. This way you can get variables in
-            templates by using ``{{ var.json.variable_name }}`` or
-            ``{{ var.json.get('variable_name', {'fall': 'back'}) }}``.
-            """
-
-            def __init__(self):
-                self.var = None
-
-            def __getattr__(
-                self,
-                item: str,
-            ):
-                self.var = Variable.get(item, deserialize_json=True)
-                return self.var
-
-            def __repr__(self):
-                return str(self.var)
-
-            @staticmethod
-            def get(
-                item: str,
-                default_var: Any = Variable._Variable__NO_DEFAULT_SENTINEL,
-            ):
-                """Get Airflow Variable after deserializing JSON value"""
-                return Variable.get(item, default_var=default_var, deserialize_json=True)
-
-        class ConnectionAccessor:
-            """
-            Wrapper around Connection. This way you can get connections in
-            templates by using ``{{ conn.conn_id }}`` or
-            ``{{ conn.get('conn_id') }}``.
-            """
-
-            def __getattr__(
-                self,
-                item: str,
-            ):
-                return Connection.get_connection_from_secrets(item)
-
-            @staticmethod
-            def get(
-                item: str,
-                default_conn: Any = None,
-            ):
-                """Get Airflow Connection value"""
-                try:
-                    return Connection.get_connection_from_secrets(item)
-                except AirflowNotFoundException:
-                    return default_conn
-
-        # Create lazy proxies for deprecated stuff.
-
-        def deprecated_proxy(
-            func: Callable[[], Any],
-            *,
-            key: str,
-            replacements: Optional[List[str]] = None,
-        ) -> lazy_object_proxy.Proxy:
-            def deprecated_func():
-                message = (
-                    f"Accessing {key!r} from the template is deprecated and "
-                    f"will be removed in a future version."
-                )
-                if replacements:
-                    display_except_last = ", ".join(repr(r) for r in replacements[:-1])
-                    if display_except_last:
-                        message += f" Please use {display_except_last} or {replacements[-1]!r} instead."
-                    else:
-                        message += f" Please use {replacements[-1]!r} instead."
-                warnings.warn(message, DeprecationWarning)
-                return func()
-
-            return lazy_object_proxy.Proxy(deprecated_func)
-
         @cache
         def get_yesterday_ds() -> str:
-            return (self.execution_date - timedelta(1)).strftime('%Y-%m-%d')
+            return (logical_date - timedelta(1)).strftime('%Y-%m-%d')
 
         def get_yesterday_ds_nodash() -> str:
             return get_yesterday_ds().replace('-', '')
 
         @cache
         def get_tomorrow_ds() -> str:
-            return (self.execution_date + timedelta(1)).strftime('%Y-%m-%d')
+            return (logical_date + timedelta(1)).strftime('%Y-%m-%d')
 
         def get_tomorrow_ds_nodash() -> str:
             return get_tomorrow_ds().replace('-', '')
@@ -1960,18 +1925,15 @@ class TaskInstance(Base, LoggingMixin):
         @cache
         def get_next_execution_date() -> Optional[pendulum.DateTime]:
             # For manually triggered dagruns that aren't run on a schedule,
-            # next/previous execution dates don't make sense, and should be set
+            # the "next" execution date doesn't make sense, and should be set
             # to execution date for consistency with how execution_date is set
             # for manually triggered tasks, i.e. triggered_date == execution_date.
             if dag_run.external_trigger:
-                next_execution_date = dag_run.execution_date
-            else:
-                with warnings.catch_warnings():
-                    warnings.simplefilter("ignore", DeprecationWarning)
-                    next_execution_date = dag.following_schedule(self.execution_date)
-            if next_execution_date is None:
+                return logical_date
+            next_info = dag.next_dagrun_info(data_interval, restricted=False)
+            if next_info is None:
                 return None
-            return timezone.coerce_datetime(next_execution_date)
+            return timezone.coerce_datetime(next_info.logical_date)
 
         def get_next_ds() -> Optional[str]:
             execution_date = get_next_execution_date()
@@ -1987,11 +1949,15 @@ class TaskInstance(Base, LoggingMixin):
 
         @cache
         def get_prev_execution_date():
+            # For manually triggered dagruns that aren't run on a schedule,
+            # the "previous" execution date doesn't make sense, and should be set
+            # to execution date for consistency with how execution_date is set
+            # for manually triggered tasks, i.e. triggered_date == execution_date.
             if dag_run.external_trigger:
-                return timezone.coerce_datetime(self.execution_date)
+                return logical_date
             with warnings.catch_warnings():
                 warnings.simplefilter("ignore", DeprecationWarning)
-                return dag.previous_schedule(self.execution_date)
+                return dag.previous_schedule(logical_date)
 
         @cache
         def get_prev_ds() -> Optional[str]:
@@ -2006,7 +1972,7 @@ class TaskInstance(Base, LoggingMixin):
                 return None
             return prev_ds.replace('-', '')
 
-        return {
+        context = {
             'conf': conf,
             'dag': dag,
             'dag_run': dag_run,
@@ -2014,57 +1980,45 @@ class TaskInstance(Base, LoggingMixin):
             'data_interval_start': timezone.coerce_datetime(data_interval.start),
             'ds': ds,
             'ds_nodash': ds_nodash,
-            'execution_date': deprecated_proxy(
-                lambda: logical_date,
-                key='execution_date',
-                replacements=['logical_date', 'data_interval_start'],
-            ),
+            'execution_date': logical_date,
             'inlets': task.inlets,
             'logical_date': logical_date,
             'macros': macros,
-            'next_ds': deprecated_proxy(get_next_ds, key="next_ds", replacements=["data_interval_end | ds"]),
-            'next_ds_nodash': deprecated_proxy(
-                get_next_ds_nodash,
-                key="next_ds_nodash",
-                replacements=["data_interval_end | ds_nodash"],
-            ),
-            'next_execution_date': deprecated_proxy(
-                get_next_execution_date,
-                key='next_execution_date',
-                replacements=['data_interval_end'],
-            ),
+            'next_ds': get_next_ds(),
+            'next_ds_nodash': get_next_ds_nodash(),
+            'next_execution_date': get_next_execution_date(),
             'outlets': task.outlets,
             'params': task.params,
-            'prev_data_interval_start_success': lazy_object_proxy.Proxy(get_prev_data_interval_start_success),
-            'prev_data_interval_end_success': lazy_object_proxy.Proxy(get_prev_data_interval_end_success),
-            'prev_ds': deprecated_proxy(get_prev_ds, key="prev_ds"),
-            'prev_ds_nodash': deprecated_proxy(get_prev_ds_nodash, key="prev_ds_nodash"),
-            'prev_execution_date': deprecated_proxy(get_prev_execution_date, key='prev_execution_date'),
-            'prev_execution_date_success': deprecated_proxy(
-                lambda: self.get_previous_execution_date(state=State.SUCCESS, session=session),
-                key='prev_execution_date_success',
-                replacements=['prev_data_interval_start_success'],
+            'prev_data_interval_start_success': get_prev_data_interval_start_success(),
+            'prev_data_interval_end_success': get_prev_data_interval_end_success(),
+            'prev_ds': get_prev_ds(),
+            'prev_ds_nodash': get_prev_ds_nodash(),
+            'prev_execution_date': get_prev_execution_date(),
+            'prev_execution_date_success': self.get_previous_execution_date(
+                state=State.SUCCESS,
+                session=session,
             ),
-            'prev_start_date_success': lazy_object_proxy.Proxy(get_prev_start_date_success),
+            'prev_start_date_success': get_prev_start_date_success(),
             'run_id': self.run_id,
             'task': task,
             'task_instance': self,
             'task_instance_key_str': f"{task.dag_id}__{task.task_id}__{ds_nodash}",
             'test_mode': self.test_mode,
             'ti': self,
-            'tomorrow_ds': deprecated_proxy(get_tomorrow_ds, key='tomorrow_ds'),
-            'tomorrow_ds_nodash': deprecated_proxy(get_tomorrow_ds_nodash, key='tomorrow_ds_nodash'),
+            'tomorrow_ds': get_tomorrow_ds(),
+            'tomorrow_ds_nodash': get_tomorrow_ds_nodash(),
             'ts': ts,
             'ts_nodash': ts_nodash,
             'ts_nodash_with_tz': ts_nodash_with_tz,
             'var': {
-                'json': VariableJsonAccessor(),
-                'value': VariableAccessor(),
+                'json': VariableAccessor(deserialize_json=True),
+                'value': VariableAccessor(deserialize_json=False),
             },
             'conn': ConnectionAccessor(),
-            'yesterday_ds': deprecated_proxy(get_yesterday_ds, key='yesterday_ds'),
-            'yesterday_ds_nodash': deprecated_proxy(get_yesterday_ds_nodash, key='yesterday_ds_nodash'),
+            'yesterday_ds': get_yesterday_ds(),
+            'yesterday_ds_nodash': get_yesterday_ds_nodash(),
         }
+        return Context(context)
 
     @provide_session
     def get_rendered_template_fields(self, session=None):
@@ -2134,7 +2088,7 @@ class TaskInstance(Base, LoggingMixin):
         sanitized_pod = ApiClient().sanitize_for_serialization(pod)
         return sanitized_pod
 
-    def get_email_subject_content(self, exception):
+    def get_email_subject_content(self, exception: BaseException) -> Tuple[str, str, str]:
         """Get the email subject content for exceptions."""
         # For a ti from DB (without ti.task), return the default value
         # Reuse it for smart sensor to send default email alert
@@ -2161,18 +2115,18 @@ class TaskInstance(Base, LoggingMixin):
             'Mark success: <a href="{{ti.mark_success_url}}">Link</a><br>'
         )
 
+        # This function is called after changing the state from State.RUNNING,
+        # so we need to subtract 1 from self.try_number here.
+        current_try_number = self.try_number - 1
+        additional_context = {
+            "exception": exception,
+            "exception_html": exception_html,
+            "try_number": current_try_number,
+            "max_tries": self.max_tries,
+        }
+
         if use_default:
-            jinja_context = {'ti': self}
-            # This function is called after changing the state
-            # from State.RUNNING so need to subtract 1 from self.try_number.
-            jinja_context.update(
-                dict(
-                    exception=exception,
-                    exception_html=exception_html,
-                    try_number=self.try_number - 1,
-                    max_tries=self.max_tries,
-                )
-            )
+            jinja_context = {"ti": self, **additional_context}
             jinja_env = jinja2.Environment(
                 loader=jinja2.FileSystemLoader(os.path.dirname(__file__)), autoescape=True
             )
@@ -2182,24 +2136,15 @@ class TaskInstance(Base, LoggingMixin):
 
         else:
             jinja_context = self.get_template_context()
-
-            jinja_context.update(
-                dict(
-                    exception=exception,
-                    exception_html=exception_html,
-                    try_number=self.try_number - 1,
-                    max_tries=self.max_tries,
-                )
-            )
-
+            jinja_context.update(additional_context)
             jinja_env = self.task.get_template_env()
 
-            def render(key, content):
+            def render(key: str, content: str) -> str:
                 if conf.has_option('email', key):
                     path = conf.get('email', key)
                     with open(path) as f:
                         content = f.read()
-                return jinja_env.from_string(content).render(**jinja_context)
+                return render_template_to_string(jinja_env.from_string(content), jinja_context)
 
             subject = render('subject_template', default_subject)
             html_content = render('html_content_template', default_html_content)
